from typing import Any, Dict, List

def handler_decorator(d: Dict[str, List[Any]]):
    def _handler(command: str):
        def _(func: Any):
            if not command in d:
                d[command] = []
            d[command].append(func)
            return func
        return _
    return _handler



import string
from typing import List

ASCII_UPPER   = list(string.ascii_uppercase)
ASCII_LOWER   = list(string.ascii_lowercase)
RFC1459_UPPER = ASCII_UPPER+list("[]^\\")
RFC1459_LOWER = ASCII_LOWER+list("{}~|")

def _replace(s: str, upper: List[str], lower: List[str]):
    out = ""
    for char in s:
        if char in upper:
            out += lower[upper.index(char)]
        else:
            out += char
    return out

def casefold(mapping: str, s: str):
    if mapping   == "rfc1459":
        return _replace(s, RFC1459_UPPER, RFC1459_LOWER)
    elif mapping == "ascii":
        return _replace(s, ASCII_UPPER,   ASCII_LOWER)



import unittest
import ircstates, irctokens

class ISUPPORTTest(unittest.TestCase):
    def test_chanmodes(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.chanmodes.a_modes, ["b"])
        self.assertEqual(server.isupport.chanmodes.b_modes, ["k"])
        self.assertEqual(server.isupport.chanmodes.c_modes, ["l"])
        self.assertEqual(server.isupport.chanmodes.d_modes,
            ["i", "m", "n", "p", "s", "t"])
        server.parse_tokens(irctokens.tokenise("005 * CHANMODES=a,b,c,d *"))
        self.assertEqual(server.isupport.chanmodes.a_modes, ["a"])
        self.assertEqual(server.isupport.chanmodes.b_modes, ["b"])
        self.assertEqual(server.isupport.chanmodes.c_modes, ["c"])
        self.assertEqual(server.isupport.chanmodes.d_modes, ["d"])

    def test_prefix(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.prefix.modes, ["o", "v"])
        self.assertEqual(server.isupport.prefix.prefixes, ["@", "+"])

        self.assertEqual(server.isupport.prefix.from_mode("o"), "@")
        self.assertIsNone(server.isupport.prefix.from_mode("a"))
        self.assertEqual(server.isupport.prefix.from_prefix("@"), "o")
        self.assertIsNone(server.isupport.prefix.from_prefix("&"))

        server.parse_tokens(irctokens.tokenise("005 * PREFIX=(qaohv)~&@%+ *"))
        self.assertEqual(server.isupport.prefix.modes,
            ["q", "a", "o", "h", "v"])
        self.assertEqual(server.isupport.prefix.prefixes,
            ["~", "&", "@", "%", "+"])
        self.assertEqual(server.isupport.prefix.from_mode("a"), "&")
        self.assertEqual(server.isupport.prefix.from_prefix("&"), "a")


    def test_chantypes(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.chantypes, ["#"])
        server.parse_tokens(irctokens.tokenise("005 * CHANTYPES=#& *"))
        self.assertEqual(server.isupport.chantypes, ["#", "&"])

    def test_modes(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.modes, 3)
        server.parse_tokens(irctokens.tokenise("005 * MODES *"))
        self.assertEqual(server.isupport.modes, -1)
        server.parse_tokens(irctokens.tokenise("005 * MODES=5 *"))
        self.assertEqual(server.isupport.modes, 5)

    def test_rfc1459(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.casemapping, "rfc1459")
        server.parse_tokens(irctokens.tokenise("005 * CASEMAPPING=rfc1459 *"))
        self.assertEqual(server.isupport.casemapping, "rfc1459")

    def test_ascii(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("005 * CASEMAPPING=ascii *"))
        self.assertEqual(server.isupport.casemapping, "ascii")

    def test_fallback_to_rfc1459(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("005 * CASEMAPPING=asd *"))
        self.assertEqual(server.isupport.casemapping, "rfc1459")

    def test_network(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.network)
        server.parse_tokens(irctokens.tokenise("005 * NETWORK=testnet *"))
        self.assertEqual(server.isupport.network, "testnet")

    def test_statusmsg(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertEqual(server.isupport.statusmsg, [])
        server.parse_tokens(irctokens.tokenise("005 * STATUSMSG=&@ *"))
        self.assertEqual(server.isupport.statusmsg, ["&", "@"])

    def test_callerid(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.callerid)
        server.parse_tokens(irctokens.tokenise("005 * CALLERID=U *"))
        self.assertEqual(server.isupport.callerid, "U")
        server.parse_tokens(irctokens.tokenise("005 * CALLERID *"))
        self.assertEqual(server.isupport.callerid, "g")

    def test_excepts(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.excepts)
        server.parse_tokens(irctokens.tokenise("005 * EXCEPTS=U *"))
        self.assertEqual(server.isupport.excepts, "U")
        server.parse_tokens(irctokens.tokenise("005 * EXCEPTS *"))
        self.assertEqual(server.isupport.excepts, "e")

    def test_invex(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.invex)
        server.parse_tokens(irctokens.tokenise("005 * INVEX=U *"))
        self.assertEqual(server.isupport.invex, "U")
        server.parse_tokens(irctokens.tokenise("005 * INVEX *"))
        self.assertEqual(server.isupport.invex, "I")

    def test_whox(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertFalse(server.isupport.whox)
        server.parse_tokens(irctokens.tokenise("005 * WHOX *"))
        self.assertTrue(server.isupport.whox)

    def test_monitor(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.monitor)
        server.parse_tokens(irctokens.tokenise("005 * MONITOR=123 *"))
        self.assertEqual(server.isupport.monitor, 123)
        server.parse_tokens(irctokens.tokenise("005 * MONITOR *"))
        self.assertEqual(server.isupport.monitor, -1)

    def test_watch(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        self.assertIsNone(server.isupport.watch)
        server.parse_tokens(irctokens.tokenise("005 * WATCH=123 *"))
        self.assertEqual(server.isupport.watch, 123)
        server.parse_tokens(irctokens.tokenise("005 * WATCH *"))
        self.assertEqual(server.isupport.watch, -1)



from typing import Dict, List, Optional
from .tokens import ChanModes, Prefix

CASEMAPPINGS = ["rfc1459", "ascii"]

class ISupport(object):
    raw: Dict[str, Optional[str]]

    network: Optional[str] = None

    chanmodes = ChanModes(["b"], ["k"], ["l"], ["i", "m", "n", "p", "s", "t"])
    prefix    = Prefix(["o", "v"], ["@", "+"])

    modes:       int       = 3 # -1 if "no limit"
    casemapping: str       = "rfc1459"
    chantypes:   List[str] = ["#"]
    statusmsg:   List[str] = []

    callerid: Optional[str] = None
    excepts:  Optional[str] = None
    invex:    Optional[str] = None

    monitor: Optional[int] = None # -1 if "no limit"
    watch:   Optional[int] = None # -1 if "no limit"
    whox:    bool          = False

    def __init__(self):
        self.raw = {}

    def from_tokens(self, tokens: List[str]):
        for token in tokens:
            key, sep, value = token.partition("=")
            self.raw[key] = value if sep else None

            if   key == "NETWORK":
                self.network = value

            elif key == "CHANMODES":
                a, b, c, d = value.split(",")
                self.chanmodes = ChanModes(list(a), list(b), list(c), list(d))

            elif key == "PREFIX":
                modes, prefixes = value[1:].split(")")
                self.prefix = Prefix(list(modes), list(prefixes))

            elif key == "STATUSMSG":
                self.statusmsg = list(value)

            elif key == "MODES":
                self.modes   = int(value) if value else -1
            elif key == "MONITOR":
                self.monitor = int(value) if value else -1
            elif key == "WATCH":
                self.watch   = int(value) if value else -1

            elif key == "CASEMAPPING":
                if value in CASEMAPPINGS:
                    self.casemapping = value

            elif key == "CHANTYPES":
                self.chantypes = list(value)

            elif key == "CALLERID":
                self.callerid = value or "g"
            elif key == "EXCEPTS":
                self.excepts  = value or "e"
            elif key == "INVEX":
                self.invex    = value or "I"

            elif key == "WHOX":
                self.whox = True



import unittest
import pendulum
import ircstates, irctokens

class ChannelTestJoin(unittest.TestCase):
    def test_self_join(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        self.assertIn("#chan", server.channels)
        self.assertIn("nickname", server.users)
        self.assertEqual(len(server.users), 1)
        self.assertEqual(len(server.channels), 1)

        user = server.users["nickname"]
        channel = server.channels["#chan"]
        self.assertIn(user.nickname_lower, channel.users)
        channel_user = channel.users[user.nickname_lower]
        self.assertEqual(user.channels, set([channel.name_lower]))

    def test_other_join(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        self.assertEqual(len(server.users), 2)
        self.assertIn("other", server.users)
        channel = server.channels["#chan"]
        self.assertEqual(len(channel.users), 2)

        user = server.users["other"]
        self.assertEqual(user.channels, set([channel.name_lower]))

class ChannelTestPart(unittest.TestCase):
    def test_self_part(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":nickname PART #chan"))
        self.assertEqual(len(server.users), 0)
        self.assertEqual(len(server.channels), 0)

    def test_other_part(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other PART #chan"))

        user = server.users["nickname"]
        channel = server.channels["#chan"]
        channel_user = channel.users[user.nickname_lower]

        self.assertEqual(server.users, {"nickname": user})
        self.assertEqual(server.channels, {"#chan": channel})
        self.assertEqual(user.channels, set([channel.name_lower]))
        self.assertEqual(channel.users, {"nickname": channel_user})

class ChannelTestKick(unittest.TestCase):
    def test_self_kick(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":nickname KICK #chan nickname"))
        self.assertEqual(len(server.users), 0)
        self.assertEqual(len(server.channels), 0)

    def test_other_kick(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":nickname KICK #chan other"))

        user = server.users["nickname"]
        channel = server.channels["#chan"]
        channel_user = channel.users[user.nickname_lower]

        self.assertEqual(len(server.users), 1)
        self.assertEqual(len(server.channels), 1)
        self.assertEqual(user.channels, set([channel.name_lower]))
        self.assertEqual(channel.users, {user.nickname_lower: channel_user})

class ChannelTestTopic(unittest.TestCase):
    def test_text(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("332 * #chan :test"))
        self.assertEqual(server.channels["#chan"].topic, "test")

    def test_set_by_at(self):
        dt = pendulum.datetime(2020, 3, 12, 14, 27, 57)
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("333 * #chan other 1584023277"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.topic_setter, "other")
        self.assertEqual(channel.topic_time, dt)

    def test_topic_command(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("TOPIC #chan :hello there"))
        self.assertEqual(server.channels["#chan"].topic, "hello there")

class ChannelTestCreation(unittest.TestCase):
    def test(self):
        dt = pendulum.datetime(2020, 3, 12, 19, 38, 9)
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("329 * #chan 1584041889"))
        self.assertEqual(server.channels["#chan"].created, dt)

class ChannelTestNAMES(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("353 * * #chan :nickname @+other"))
        self.assertIn("nickname", server.users)
        self.assertIn("other", server.users)

        user = server.users["other"]
        channel = server.channels["#chan"]
        channel_user_1 = channel.users[user.nickname_lower]
        channel_user_2 = channel.users[server.nickname_lower]

        self.assertEqual(channel.users, {
            user.nickname_lower: channel_user_1,
            server.nickname_lower: channel_user_2})
        self.assertEqual(user.channels, set([channel.name_lower]))
        self.assertEqual(channel_user_1.modes, ["o", "v"])

    def test_userhost_in_names(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(
            "353 * * #chan :nickname!user@host other!user2@host2"))
        self.assertEqual(server.username, "user")
        self.assertEqual(server.hostname, "host")
        user = server.users["other"]
        self.assertEqual(user.username, "user2")
        self.assertEqual(user.hostname, "host2")

class ChannelNICKAfterJoin(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        user = server.users["nickname"]
        channel = server.channels["#chan"]
        channel_user = channel.users[user.nickname_lower]
        server.parse_tokens(irctokens.tokenise(":nickname NICK Nickname2"))

        self.assertEqual(channel.users, {user.nickname_lower: channel_user})
        self.assertEqual(channel_user.nickname,       "Nickname2")
        self.assertEqual(channel_user.nickname_lower, "nickname2")



import unittest
import ircstates, irctokens


class ModeTestUMode(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("MODE nickname +i"))
        self.assertEqual(server.modes, ["i"])

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("MODE nickname +i"))
        server.parse_tokens(irctokens.tokenise("MODE nickname -i"))
        self.assertEqual(server.modes, [])

class ModeTestChannelPrefix(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("MODE #chan +ov nickname nickname"))
        user = server.users["nickname"]
        channel = server.channels["#chan"]
        channel_user = channel.users[user.nickname_lower]
        self.assertEqual(channel_user.modes, ["o", "v"])

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("MODE #chan +ov nickname nickname"))
        server.parse_tokens(
            irctokens.tokenise("MODE #chan -ov nickname nickname"))
        user = server.users["nickname"]
        channel = server.channels["#chan"]
        channel_user = channel.users[user.nickname_lower]
        self.assertEqual(channel_user.modes, [])

class ModeTestChannelList(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        channel = server.channels["#chan"]
        self.assertEqual(channel.list_modes, {"b": []})

        server.parse_tokens(irctokens.tokenise("MODE #chan +b asd!*@*"))
        self.assertEqual(channel.list_modes, {"b": ["asd!*@*"]})

        server.parse_tokens(irctokens.tokenise("MODE #chan -b asd!*@*"))
        self.assertEqual(channel.list_modes, {"b": []})

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +b asd!*@*"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +b dsa!*@*"))
        server.parse_tokens(irctokens.tokenise("MODE #chan -b asd!*@*"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.list_modes, {"b": ["dsa!*@*"]})

    def test_banlist(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("367 * #chan *!*@host setby 1594477713"))
        server.parse_tokens(
            irctokens.tokenise("367 * #chan $a:account setby 1594477713"))
        server.parse_tokens(
            irctokens.tokenise("367 * #chan r:my*gecos"))
        server.parse_tokens(irctokens.tokenise("368 * #chan *"))

        channel = server.channels["#chan"]
        self.assertEqual(
            channel.list_modes["b"],
            ["*!*@host", "$a:account", "r:my*gecos"]
        )

    def test_quietlist(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("728 * #chan q q!*@host setby 1594477713"))
        server.parse_tokens(
            irctokens.tokenise("728 * #chan q $a:qaccount setby 1594477713"))
        server.parse_tokens(
            irctokens.tokenise("728 * #chan q r:q*my*gecos setby 1594477713"))
        server.parse_tokens(irctokens.tokenise("729 * #chan q *"))

        channel = server.channels["#chan"]
        self.assertEqual(
            channel.list_modes["q"],
            ["q!*@host", "$a:qaccount", "r:q*my*gecos"]
        )


class ModeTestChannelTypeB(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +k password"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {"k": "password"})

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +k password"))
        server.parse_tokens(irctokens.tokenise("MODE #chan -k *"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {})

class ModeTestChannelTypeC(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +l 100"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {"l": "100"})

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +l 100"))
        server.parse_tokens(irctokens.tokenise("MODE #chan -l"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {})

class ModeTestChannelTypeD(unittest.TestCase):
    def test_add(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +i"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {"i": None})

    def test_remove(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("MODE #chan +i"))
        server.parse_tokens(irctokens.tokenise("MODE #chan -i"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {})

class ModeTestChannelNumeric(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("324 * #chan +bkli *!*@* pass 10"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {"k": "pass", "l": "10", "i": None})
        self.assertEqual(channel.list_modes, {"b": ["*!*@*"]})

    def test_without_plus(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("324 * #chan il 10"))
        channel = server.channels["#chan"]
        self.assertEqual(channel.modes, {"i": None, "l": "10"})

class ModeTestUserNumeric(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("221 * +iw"))
        self.assertEqual(server.modes, ["i", "w"])

    def test_without_plus(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("221 * iw"))
        self.assertEqual(server.modes, ["i", "w"])



import unittest
import ircstates, irctokens

class UserTestNicknameChange(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname NICK Nickname2"))
        self.assertEqual(server.nickname,       "Nickname2")
        self.assertEqual(server.nickname_lower, "nickname2")

        server.parse_tokens(irctokens.tokenise(":nickname2 JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        self.assertIn("other", server.users)
        user = server.users["other"]

        server.parse_tokens(irctokens.tokenise(":other NICK Other2"))
        self.assertNotIn("other", server.users)
        self.assertIn("other2", server.users)

        self.assertEqual(user.nickname,       "Other2")
        self.assertEqual(user.nickname_lower, "other2")

class UserTestHostmaskJoin(unittest.TestCase):
    def test_both(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(
            irctokens.tokenise(":nickname!user@host JOIN #chan"))
        self.assertEqual(server.username, "user")
        self.assertEqual(server.hostname, "host")

        server.parse_tokens(irctokens.tokenise(":other!user@host JOIN #chan"))
        user = server.users["other"]
        self.assertEqual(user.username, "user")
        self.assertEqual(user.hostname, "host")
        self.assertEqual(user.userhost(), "user@host")
        self.assertEqual(user.hostmask(), "other!user@host")

    def test_user(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname!user JOIN #chan"))
        self.assertEqual(server.username, "user")
        self.assertIsNone(server.hostname)

        server.parse_tokens(irctokens.tokenise(":other!user JOIN #chan"))
        user = server.users["other"]
        self.assertEqual(user.username, "user")
        self.assertIsNone(user.hostname)
        self.assertIsNone(user.userhost())
        self.assertEqual(user.hostmask(), "other!user")

    def test_host(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname@host JOIN #chan"))
        self.assertIsNone(server.username)
        self.assertEqual(server.hostname, "host")

        server.parse_tokens(irctokens.tokenise(":other@host JOIN #chan"))
        user = server.users["other"]
        self.assertIsNone(user.username)
        self.assertEqual(user.hostname, "host")
        self.assertIsNone(user.userhost())
        self.assertEqual(user.hostmask(), "other@host")

class UserTestExtendedJoin(unittest.TestCase):
    def test_without_extended_join(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        self.assertIsNone(server.account)
        self.assertIsNone(server.realname)
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        user = server.users["other"]
        self.assertIsNone(user.account)
        self.assertIsNone(user.realname)

    def test_with_account(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan acc :realname"))
        self.assertEqual(server.account, "acc")
        self.assertEqual(server.realname, "realname")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan acc2 :realname2"))
        user = server.users["other"]
        self.assertEqual(user.account, "acc2")
        self.assertEqual(user.realname, "realname2")

    def test_without_account(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan * :realname"))
        self.assertEqual(server.account, "")
        self.assertEqual(server.realname, "realname")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan * :realname2"))
        user = server.users["other"]
        self.assertEqual(user.account, "")
        self.assertEqual(user.realname, "realname2")

class UserTestAccountNotify(unittest.TestCase):
    def test_with_account(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":nickname ACCOUNT acc"))
        self.assertEqual(server.account, "acc")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other ACCOUNT acc2"))
        user = server.users["other"]
        self.assertEqual(user.account, "acc2")

    def test_without_account(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":nickname ACCOUNT *"))
        self.assertEqual(server.account, "")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other ACCOUNT *"))
        user = server.users["other"]
        self.assertEqual(user.account, "")

class UserTestHostmaskPRIVMSG(unittest.TestCase):
    def test_both(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":nickname!user@host PRIVMSG #chan :hi"))
        self.assertEqual(server.username, "user")
        self.assertEqual(server.hostname, "host")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":other!user@host PRIVMSG #chan :hi"))
        user = server.users["other"]
        self.assertEqual(user.username, "user")
        self.assertEqual(user.hostname, "host")

    def test_user(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":nickname!user PRIVMSG #chan :hi"))
        self.assertEqual(server.username, "user")
        self.assertIsNone(server.hostname)
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":other!user PRIVMSG #chan :hi"))
        user = server.users["other"]
        self.assertEqual(user.username, "user")
        self.assertIsNone(user.hostname)

    def test_host(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":nickname@host PRIVMSG #chan :hi"))
        self.assertIsNone(server.username)
        self.assertEqual(server.hostname, "host")
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise(":other@host PRIVMSG #chan :hi"))
        user = server.users["other"]
        self.assertIsNone(user.username)
        self.assertEqual(user.hostname, "host")

class UserTestVisibleHost(unittest.TestCase):
    def test_without_username(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("396 * hostname *"))
        self.assertIsNone(server.username)
        self.assertEqual(server.hostname, "hostname")

    def test_with_username(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("396 * username@hostname *"))
        self.assertEqual(server.username, "username")
        self.assertEqual(server.hostname, "hostname")

class UserTestWHO(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(
            irctokens.tokenise("352 * #chan user host * nickname * :0 real"))
        server.parse_tokens(
            irctokens.tokenise("352 * #chan user2 host2 * other * :0 real2"))

        self.assertEqual(server.username, "user")
        self.assertEqual(server.hostname, "host")
        self.assertEqual(server.realname, "real")
        user = server.users["other"]
        self.assertEqual(user.username, "user2")
        self.assertEqual(user.hostname, "host2")
        self.assertEqual(user.realname, "real2")

class UserTestCHGHOST(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(
            irctokens.tokenise(":nickname!user@host JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":nickname CHGHOST u h"))
        self.assertEqual(server.username, "u")
        self.assertEqual(server.hostname, "h")
        server.parse_tokens(
            irctokens.tokenise(":other!user2@host2 JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other CHGHOST u2 h2"))
        user = server.users["other"]
        self.assertEqual(user.username, "u2")
        self.assertEqual(user.hostname, "h2")

class UserTestWHOIS(unittest.TestCase):
    def test_user_line(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("311 * nickname u h * :r"))
        self.assertEqual(server.username, "u")
        self.assertEqual(server.hostname, "h")
        self.assertEqual(server.realname, "r")

        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        server.parse_tokens(irctokens.tokenise("311 * other u2 h2 * :r2"))
        user = server.users["other"]
        self.assertEqual(user.username, "u2")
        self.assertEqual(user.hostname, "h2")
        self.assertEqual(user.realname, "r2")

class UserTestAway(unittest.TestCase):
    def test_verb_set(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        user = server.users["nickname"]
        self.assertIsNone(server.away)
        self.assertIsNone(user.away)

        server.parse_tokens(irctokens.tokenise(":nickname AWAY :ik ga weg"))
        self.assertEqual(server.away, "ik ga weg")
        self.assertEqual(user.away,   "ik ga weg")

    def test_verb_unset(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]

        server.parse_tokens(irctokens.tokenise(
            ":nickname AWAY :let's blow this popsicle stand"))
        server.parse_tokens(irctokens.tokenise(":nickname AWAY"))
        self.assertIsNone(server.away)
        self.assertIsNone(user.away)

    def test_numeric(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]

        self.assertIsNone(server.away)
        self.assertIsNone(user.away)

        server.parse_tokens(irctokens.tokenise(
            "301 * nickname :i saw something shiny"))
        self.assertEqual(server.away, "i saw something shiny")
        self.assertEqual(user.away,   "i saw something shiny")

class UserTestSETNAME(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        user = server.users["other"]
        self.assertIsNone(user.realname)
        self.assertIsNone(server.realname)
        server.parse_tokens(
            irctokens.tokenise(":nickname SETNAME :new now know how"))
        server.parse_tokens(
            irctokens.tokenise(":other SETNAME :tyrannosaurus hex"))
        self.assertEqual(server.realname, "new now know how")
        self.assertEqual(user.realname, "tyrannosaurus hex")



from ipaddress import ip_address
from typing    import Callable, Dict, List, Optional, Set, Tuple
from irctokens import Line, build, Hostmask, StatefulDecoder, StatefulEncoder
from irctokens import hostmask as hostmask_
from pendulum  import from_timestamp, now

from .user         import User
from .channel      import Channel
from .channel_user import ChannelUser
from .isupport     import ISupport
from .decorators   import handler_decorator
from .casemap      import casefold
from .names        import Name
from .emit         import *
from .numerics     import *

LINE_HANDLERS: Dict[str, List[Callable[["Server", Line], Emit]]] = {}
line_handler = handler_decorator(LINE_HANDLERS)

class ServerException(Exception):
    pass
class ServerDisconnectedException(ServerException):
    pass

WHO_TYPE = "735" # randomly generated
TYPE_EMIT = Optional[Emit]

class Server(object):
    def __init__(self, name: str):
        self.name = name

        self.nickname                = ""
        self.nickname_lower          = ""
        self.username: Optional[str] = None
        self.hostname: Optional[str] = None
        self.realname: Optional[str] = None
        self.account:  Optional[str] = None
        self.server:   Optional[str] = None
        self.away:     Optional[str] = None
        self.ip:       Optional[str] = None

        self.registered = False
        self.modes: List[str] = []
        self.motd:  List[str] = []

        self._decoder = StatefulDecoder()

        self.users:    Dict[str, User]    = {}
        self.channels: Dict[str, Channel] = {}

        self.isupport = ISupport()

        self.has_cap: bool = False
        self._temp_caps:     Dict[str, str] = {}
        self.available_caps: Dict[str, str] = {}
        self.agreed_caps:    List[str]      = []

    def __repr__(self) -> str:
        return f"Server(name={self.name!r})"

    def recv(self, data: bytes) -> List[Line]:
        lines = self._decoder.push(data)
        if lines is None:
            raise ServerDisconnectedException()
        return lines

    def parse_tokens(self, line: Line) -> TYPE_EMIT:
        ret_emit: TYPE_EMIT = None
        if line.command in LINE_HANDLERS:
            for callback in LINE_HANDLERS[line.command]:
                emit = callback(self, line)
                if emit is not None and ret_emit is None:
                    emit.command = line.command
                    ret_emit = emit
        return ret_emit

    def casefold(self, s1: str):
        return casefold(self.isupport.casemapping, s1)
    def casefold_equals(self, s1: str, s2: str):
        return self.casefold(s1) == self.casefold(s2)
    def is_me(self, nickname: str):
        return self.casefold(nickname) == self.nickname_lower

    def has_user(self, nickname: str) -> bool:
        return self.casefold(nickname) in self.users
    def _add_user(self, nickname: str, nickname_lower: str):
        user = self.create_user(Name(nickname, nickname_lower))
        self.users[nickname_lower] = user

    def is_channel(self, target: str) -> bool:
        return target[:1] in self.isupport.chantypes
    def has_channel(self, name: str) -> bool:
        return self.casefold(name) in self.channels
    def get_channel(self, name: str) -> Optional[Channel]:
        return self.channels.get(self.casefold(name), None)

    def create_user(self, nickname: Name) -> User:
        return User(nickname)

    def create_channel(self, name: Name) -> Channel:
        return Channel(name)

    def _user_join(self, channel: Channel, user: User) -> ChannelUser:
        channel_user = ChannelUser(
            user.get_name(),
            channel.get_name())

        user.channels.add(self.casefold(channel.name))
        channel.users[user.nickname_lower] = channel_user
        return channel_user

    def prepare_whox(self, target: str) -> Line:
        return build("WHO", [target, f"n%afhinrstu,{WHO_TYPE}"])

    def _self_hostmask(self, hostmask: Hostmask):
        self.nickname = hostmask.nickname
        if hostmask.username:
            self.username = hostmask.username
        if hostmask.hostname:
            self.hostname = hostmask.hostname

    def _emit(self) -> Emit:
        return Emit()

    @line_handler(RPL_WELCOME)
    # first message reliably sent to us after registration is complete
    def _handle_welcome(self, line: Line) -> Emit:
        self.nickname = line.params[0]
        self.nickname_lower = self.casefold(line.params[0])
        self.registered = True
        return self._emit()

    @line_handler(RPL_ISUPPORT)
    # https://defs.ircdocs.horse/defs/isupport.html
    def _handle_ISUPPORT(self, line: Line) -> Emit:
        self.isupport.from_tokens(line.params[1:-1])
        return self._emit()

    @line_handler(RPL_MOTDSTART)
    # start of MOTD
    def _handle_motd_start(self, line: Line) -> Emit:
        self.motd.clear()
        return self._emit()
    @line_handler(RPL_MOTDSTART)
    # start of MOTD
    @line_handler(RPL_MOTD)
    # line of MOTD
    def _handle_motd_line(self, line: Line) -> Emit:
        emit = self._emit()
        text = line.params[1]
        emit.text = text
        self.motd.append(text)
        return emit

    @line_handler("NICK")
    def _handle_NICK(self, line: Line) -> Emit:
        new_nickname       = line.params[0]
        new_nickname_lower = self.casefold(new_nickname)
        nickname_lower     = self.casefold(line.hostmask.nickname)

        emit = self._emit()

        if nickname_lower in self.users:
            user = self.users.pop(nickname_lower)
            emit.user = user
            user.change_nickname(new_nickname, new_nickname_lower)
            self.users[new_nickname_lower] = user

            for channel_lower in user.channels:
                channel = self.channels[channel_lower]
                channel_user = channel.users.pop(nickname_lower)
                channel.users[user.nickname_lower] = channel_user

        if nickname_lower == self.nickname_lower:
            emit.self = True

            self.nickname       = new_nickname
            self.nickname_lower = new_nickname_lower
        return emit

    @line_handler("JOIN")
    def _handle_JOIN(self, line: Line) -> Emit:
        extended = len(line.params) == 3

        account = line.params[1].strip("*") if extended else None
        realname = line.params[2] if extended else None

        emit = self._emit()

        channel_lower = self.casefold(line.params[0])
        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            if not channel_lower in self.channels:
                channel = self.create_channel(
                    Name(line.params[0], channel_lower)
                )
                #TODO: put this somewhere better
                for mode in self.isupport.chanmodes.a_modes:
                    channel.list_modes[mode] = []

                self.channels[channel_lower] = channel

            self._self_hostmask(line.hostmask)
            if extended:
                self.account  = account
                self.realname = realname

        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            if not nickname_lower in self.users:
                self._add_user(line.hostmask.nickname, nickname_lower)

            user = self.users[nickname_lower]
            emit.user = user
            if line.hostmask.username:
                user.username = line.hostmask.username
            if line.hostmask.hostname:
                user.hostname = line.hostmask.hostname
            if extended:
                user.account  = account
                user.realname = realname

            self._user_join(channel, user)
        return emit

    def _user_part(self, line: Line,
            nickname: str,
            channel_name: str,
            reason_i: int) -> Tuple[Emit, Optional[User]]:
        emit = self._emit()
        channel_lower = self.casefold(channel_name)
        reason = line.params[reason_i] if line.params[reason_i:] else None
        if not reason is None:
            emit.text = reason

        user: Optional[User] = None
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel

            nickname_lower = self.casefold(nickname)
            if nickname_lower in self.users:
                user = self.users[nickname_lower]

                user.channels.remove(channel.name_lower)
                del channel.users[user.nickname_lower]
                if not user.channels:
                    del self.users[nickname_lower]

            if nickname_lower == self.nickname_lower:
                del self.channels[channel_lower]

                for key, cuser in channel.users.items():
                    ruser = self.users[key]
                    ruser.channels.remove(channel.name_lower)
                    if not ruser.channels:
                        del self.users[ruser.nickname_lower]

        return emit, user

    @line_handler("PART")
    def _handle_PART(self, line: Line) -> Emit:
        emit, user = self._user_part(line, line.hostmask.nickname,
            line.params[0], 1)
        if not user is None:
            emit.user = user
            if user.nickname_lower == self.nickname_lower:
                emit.self = True
        return emit
    @line_handler("KICK")
    def _handle_KICK(self, line: Line) -> Emit:
        emit, kicked = self._user_part(line, line.params[1], line.params[0],
            2)
        if not kicked is None:
            emit.user_target = kicked

            if kicked.nickname_lower == self.nickname_lower:
                emit.self = True

            kicker_lower = self.casefold(line.hostmask.nickname)
            if kicker_lower == self.nickname_lower:
                emit.self_source = True

            if kicker_lower in self.users:
                emit.user_source = self.users[kicker_lower]
            else:
                emit.user_source = self.create_user(
                    Name(line.hostmask.nickname, kicker_lower)
                )

        return emit

    def _self_quit(self):
        self.users.clear()
        self.channels.clear()

    @line_handler("QUIT")
    def _handle_quit(self, line: Line) -> Emit:
        emit = self._emit()
        nickname_lower = self.casefold(line.hostmask.nickname)
        reason = line.params[0] if line.params else None
        if not reason is None:
            emit.text = reason

        if nickname_lower == self.nickname_lower:
            emit.self = True
            self._self_quit()
        else:
            if nickname_lower in self.users:
                user = self.users.pop(nickname_lower)
                emit.user = user
                for channel_lower in user.channels:
                    channel = self.channels[channel_lower]
                    del channel.users[user.nickname_lower]
        return emit

    @line_handler("ERROR")
    def _handle_ERROR(self, line: Line) -> Emit:
        self._self_quit()
        return self._emit()

    @line_handler(RPL_NAMREPLY)
    # channel's user list, "NAMES #channel" response (and on-join)
    def _handle_names(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[2])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            nicknames = list(filter(bool, line.params[3].split(" ")))
            users: List[User] = []
            emit.users = users

            for nickname in nicknames:
                modes = ""
                for char in nickname:
                    mode = self.isupport.prefix.from_prefix(char)
                    if mode:
                        modes += mode
                    else:
                        break

                hostmask = hostmask_(nickname[len(modes):])
                nickname_lower = self.casefold(hostmask.nickname)
                if not nickname_lower in self.users:
                    self._add_user(hostmask.nickname, nickname_lower)
                user = self.users[nickname_lower]
                users.append(user)
                channel_user = self._user_join(channel, user)

                if hostmask.username:
                    user.username = hostmask.username
                if hostmask.hostname:
                    user.hostname = hostmask.hostname

                if nickname_lower == self.nickname_lower:
                    self._self_hostmask(hostmask)

                for mode in modes:
                    if not mode in channel_user.modes:
                        channel_user.modes.append(mode)
        return emit

    @line_handler(RPL_CREATIONTIME)
    # channel creation time, "MODE #channel" response (and on-join)
    def _handle_creation_time(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[1])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            channel.created = from_timestamp(int(line.params[2]))
        return emit

    @line_handler("TOPIC")
    def _handle_TOPIC(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[0])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            channel.topic        = line.params[1]
            channel.topic_setter = line.source
            channel.topic_time   = now("utc")
        return emit

    @line_handler(RPL_TOPIC)
    # topic text, "TOPIC #channel" response (and on-join)
    def _handle_topic_num(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[1])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            self.channels[channel_lower].topic = line.params[2]
        return emit
    @line_handler(RPL_TOPICWHOTIME)
    # topic setby, "TOPIC #channel" response (and on-join)
    def _handle_topic_time(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[1])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            channel.topic_setter = line.params[2]
            channel.topic_time   = from_timestamp(int(line.params[3]))
        return emit

    def _channel_modes(self,
            channel: Channel,
            modes: List[str],
            params: List[str]
            ) -> List[Tuple[str, Optional[str]]]:
        tokens: List[Tuple[str, Optional[str]]] = []

        for mode in modes:
            add  = mode[0] == "+"
            char = mode[1]
            arg: Optional[str] = None

            if char in self.isupport.prefix.modes: # a user's status
                arg = params.pop(0)
                nickname_lower = self.casefold(arg)

                if nickname_lower in self.users:
                    user = self.users[nickname_lower]
                    channel_user = channel.users[user.nickname_lower]
                    if add:
                        if not char in channel_user.modes:
                            channel_user.modes.append(char)
                    elif char in channel_user.modes:
                        channel_user.modes.remove(char)
            else:
                has_arg = False
                is_list = False
                if char in self.isupport.chanmodes.a_modes:
                    has_arg = True
                    is_list = True
                elif add:
                    has_arg = char in (self.isupport.chanmodes.b_modes+
                        self.isupport.chanmodes.c_modes)
                else: # remove
                    has_arg = char in self.isupport.chanmodes.b_modes

                if has_arg:
                    arg = params.pop(0)

                if add:
                    channel.add_mode(char, arg, is_list)
                else:
                    channel.remove_mode(char, arg)

            tokens.append((mode, arg))

        return tokens

    @line_handler("MODE")
    def _handle_MODE(self, line: Line) -> Emit:
        emit = self._emit()
        target     = line.params[0]
        modes_str  = line.params[1]
        params     = line.params[2:].copy()

        modifier          = "+"
        modes: List[str]  = []

        for c in list(modes_str):
            if c in ["+", "-"]:
                modifier = c
            else:
                modes.append(f"{modifier}{c}")

        target_lower = self.casefold(target)
        if target_lower == self.nickname_lower:
            emit.self_target = True
            emit.tokens = modes

            for mode in modes:
                add  = mode[0] == "+"
                char = mode[1]
                if add:
                    if not char in self.modes:
                        self.modes.append(char)
                elif char in self.modes:
                    self.modes.remove(char)
        elif target_lower in self.channels:
            channel = self.channels[self.casefold(target)]
            emit.channel = channel
            ctokens = self._channel_modes(channel, modes, params)

            ctokens_str: List[str] = []
            for mode, arg in ctokens:
                if arg is not None:
                    ctokens_str.append(f"{mode} {arg}")
                else:
                    ctokens_str.append(mode)
            emit.tokens = ctokens_str
        return emit

    @line_handler(RPL_CHANNELMODEIS)
    # channel modes, "MODE #channel" response (sometimes on-join?)
    def _handle_channelmodeis(self, line: Line) -> Emit:
        emit = self._emit()
        channel_lower = self.casefold(line.params[1])
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            emit.channel = channel
            modes   = [f"+{char}" for char in line.params[2].lstrip("+")]
            params  = line.params[3:]
            self._channel_modes(channel, modes, params)
        return emit

    @line_handler(RPL_UMODEIS)
    # our own user modes, "MODE nickname" response (sometimes on-connect?)
    def _handle_umodeis(self, line: Line) -> Emit:
        for char in line.params[1].lstrip("+"):
            if not char in self.modes:
                self.modes.append(char)
        return self._emit()

    def _mode_list(self,
            channel_name: str,
            mode: str,
            mask: str):
        channel_lower = self.casefold(channel_name)
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            if not mode in channel._list_modes_temp:
                channel._list_modes_temp[mode] = []
            channel._list_modes_temp[mode].append(mask)
    def _mode_list_end(self,
            channel_name: str,
            mode: str):
        channel_lower = self.casefold(channel_name)
        if channel_lower in self.channels:
            channel = self.channels[channel_lower]
            if mode in channel._list_modes_temp:
                mlist = channel._list_modes_temp.pop(mode)
                channel.list_modes[mode] = mlist

    @line_handler(RPL_BANLIST)
    def _handle_banlist(self, line: Line) -> Emit:
        channel = line.params[1]
        mask    = line.params[2]

        if len(line.params) > 3:
            # parse these out but we're not storing them yet
            set_by  = line.params[3]
            set_at  = int(line.params[4])

        self._mode_list(channel, "b", mask)
        return self._emit()

    @line_handler(RPL_ENDOFBANLIST)
    def _handle_banlist_end(self, line: Line) -> Emit:
        channel = line.params[1]
        self._mode_list_end(channel, "b")
        return self._emit()

    @line_handler(RPL_QUIETLIST)
    def _handle_quietlist(self, line: Line) -> Emit:
        channel = line.params[1]
        mode    = line.params[2]
        mask    = line.params[3]
        set_by  = line.params[4]
        set_at  = int(line.params[5])

        self._mode_list(channel, mode, mask)
        return self._emit()

    @line_handler(RPL_ENDOFQUIETLIST)
    def _handle_quietlist_end(self, line: Line) -> Emit:
        channel = line.params[1]
        mode    = line.params[2]
        self._mode_list_end(channel, mode)
        return self._emit()

    @line_handler("PRIVMSG")
    @line_handler("NOTICE")
    @line_handler("TAGMSG")
    def _handle_message(self, line: Line) -> Emit:
        emit = self._emit()
        message = line.params[1] if line.params[1:] else None
        if not message is None:
            emit.text = message

        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self_source = True
            self._self_hostmask(line.hostmask)

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
        else:
            user = self.create_user(
                Name(line.hostmask.nickname, nickname_lower)
            )
        emit.user = user

        if line.hostmask.username:
            user.username = line.hostmask.username
        if line.hostmask.hostname:
            user.hostname = line.hostmask.hostname

        target_raw = target = line.params[0]
        statusmsg = []
        while target:
            if target[0] in self.isupport.statusmsg:
                statusmsg.append(target[0])
                target = target[1:]
            else:
                break
        emit.target = target_raw

        target_lower = self.casefold(target)
        if self.is_channel(target):
            if target_lower in self.channels:
                channel = self.channels[target_lower]
                emit.channel = channel
        elif target_lower == self.nickname_lower:
            emit.self_target = True
        return emit

    @line_handler(RPL_VISIBLEHOST)
    # our own hostname, sometimes username@hostname, when it changes
    def _handle_visiblehost(self, line: Line) -> Emit:
        username, _, hostname = line.params[1].rpartition("@")
        self.hostname = hostname
        if username:
            self.username = username
        return self._emit()

    @line_handler(RPL_WHOREPLY)
    # WHO line, "WHO #channel|nickname" response
    def _handle_who(self, line: Line) -> Emit:
        emit = self._emit()
        emit.target = line.params[1]
        nickname = line.params[5]
        username = line.params[2]
        hostname = line.params[3]
        status   = line.params[6]
        away     = "" if "G" in status else None
        realname = line.params[7].split(" ", 1)[1]

        server:  Optional[str] = None
        if not line.params[4] == "*":
            server  = line.params[4]

        nickname_lower = self.casefold(line.params[5])
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.username = username
            self.hostname = hostname
            self.realname = realname
            self.server   = server
            self.away     = away

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.username = username
            user.hostname = hostname
            user.realname = realname
            user.server   = server
            user.away     = away
        return emit

    @line_handler(RPL_WHOSPCRPL)
    # WHOX line, "WHO #channel|nickname" response; only listen for our "type"
    def _handle_whox(self, line: Line) -> Emit:
        emit = self._emit()
        if line.params[1] == WHO_TYPE and len(line.params) == 10:
            nickname_lower = self.casefold(line.params[6])
            username = line.params[2]
            hostname = line.params[4]
            status   = line.params[7]
            away     = "" if "G" in status else None
            realname = line.params[9]

            account: Optional[str] = None
            if not line.params[8] == "0":
                account = line.params[8]
            server:  Optional[str] = None
            if not line.params[5] == "*":
                server  = line.params[5]
            ip:      Optional[str] = None
            if not line.params[3] == "255.255.255.255":
                try:
                    ip = ip_address(line.params[3]).compressed
                except ValueError:
                    pass

            if nickname_lower in self.users:
                user = self.users[nickname_lower]
                emit.user = user
                user.username = username
                user.hostname = hostname
                user.realname = realname
                user.account  = account
                user.server   = server
                user.away     = away
                if ip is not None:
                    user.ip   = ip

            if nickname_lower == self.nickname_lower:
                emit.self = True
                self.username = username
                self.hostname = hostname
                self.realname = realname
                self.account  = account
                self.server   = server
                self.away     = away
                if ip is not None:
                    self.ip   = ip

        return emit

    @line_handler(RPL_WHOISUSER)
    # WHOIS "user" line, one of "WHOIS nickname" response lines
    def _handle_whoisuser(self, line: Line) -> Emit:
        emit = self._emit()
        nickname = line.params[1]
        username = line.params[2]
        hostname = line.params[3]
        realname = line.params[5]

        nickname_lower = self.casefold(nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.username = username
            self.hostname = hostname
            self.realname = realname

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.username = username
            user.hostname = hostname
            user.realname = realname
        return emit

    @line_handler("CHGHOST")
    def _handle_CHGHOST(self, line: Line) -> Emit:
        emit = self._emit()
        username = line.params[0]
        hostname = line.params[1]
        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.username = username
            self.hostname = hostname

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.username = username
            user.hostname = hostname
        return emit

    @line_handler("SETNAME")
    def _handle_SETNAME(self, line: Line) -> Emit:
        emit = self._emit()
        realname = line.params[0]
        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.realname = realname

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.realname = realname
        return emit

    @line_handler(RPL_AWAY)
    # sent in response to a command directed at a user who is marked as away
    def _handle_RPL_AWAY(self, line: Line) -> Emit:
        nickname       = line.params[1]
        nickname_lower = self.casefold(nickname)
        reason         = line.params[2]

        if nickname_lower == self.nickname_lower:
            self.away = reason
        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            user.away = reason

        return self._emit()

    @line_handler("AWAY")
    def _handle_AWAY(self, line: Line) -> Emit:
        emit = self._emit()
        away = line.params[0] if line.params else None
        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.away = away

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.away = away
        return emit

    @line_handler("ACCOUNT")
    def _handle_ACCOUNT(self, line: Line) -> Emit:
        emit = self._emit()
        account = line.params[0].strip("*")
        nickname_lower = self.casefold(line.hostmask.nickname)
        if nickname_lower == self.nickname_lower:
            emit.self = True
            self.account = account

        if nickname_lower in self.users:
            user = self.users[nickname_lower]
            emit.user = user
            user.account = account
        return emit

    @line_handler("CAP")
    def _handle_CAP(self, line: Line) -> Emit:
        self.has_cap = True
        subcommand = line.params[1].upper()
        multiline  = line.params[2] == "*"
        caps       = line.params[2 + (1 if multiline else 0)]


        tokens:     Dict[str, str] = {}
        tokens_str: List[str]      = []
        for cap in filter(bool, caps.split(" ")):
            tokens_str.append(cap)
            key, _, value = cap.partition("=")
            tokens[key] = value

        emit = self._emit()
        emit.subcommand = subcommand
        emit.finished   = not multiline
        emit.tokens     = tokens_str

        if subcommand == "LS":
            self._temp_caps.update(tokens)
            if not multiline:
                self.available_caps = self._temp_caps.copy()
                self._temp_caps.clear()
        elif subcommand == "NEW":
            self.available_caps.update(tokens)
        elif subcommand == "DEL":
            for key in tokens.keys():
                if key in self.available_caps.keys():
                    del self.available_caps[key]
                    if key in self.agreed_caps:
                        self.agreed_caps.remove(key)
        elif subcommand == "ACK":
            for key in tokens.keys():
                if key.startswith("-"):
                    key = key[1:]
                    if key in self.agreed_caps:
                        self.agreed_caps.remove(key)
                elif (not key in self.agreed_caps and
                        key in self.available_caps):
                    self.agreed_caps.append(key)
        return emit

    @line_handler(RPL_LOGGEDIN)
    def _handle_loggedin(self, line: Line) -> Emit:
        hostmask_str = line.params[1]
        hostmask     = hostmask_(hostmask_str)
        account      = line.params[2]

        self.account = account
        self._self_hostmask(hostmask)
        return self._emit()

    @line_handler(RPL_LOGGEDOUT)
    def _handle_loggedout(self, line: Line) -> Emit:
        hostmask_str = line.params[1]
        hostmask     = hostmask_(hostmask_str)

        self.account = None
        self._self_hostmask(hostmask)
        return self._emit()



from .channel  import *
from .user     import *
from .mode     import *
from .motd     import *
from .cap      import *
from .isupport import *
from .casemap  import *
from .emit     import *
from .who      import *
from .sasl     import *



from typing import List, Optional
from .user import User
from .channel import Channel

class Emit(object):
    command:    Optional[str] = None
    subcommand: Optional[str] = None

    text:   Optional[str]       = None
    tokens: Optional[List[str]] = None

    finished: Optional[bool] = None

    self:        Optional[bool] = None
    self_source: Optional[bool] = None
    self_target: Optional[bool] = None

    user:        Optional[User] = None
    user_source: Optional[User] = None
    user_target: Optional[User] = None

    users: Optional[List[User]] = None

    channel:        Optional[Channel] = None
    channel_source: Optional[Channel] = None
    channel_target: Optional[Channel] = None

    target: Optional[str] = None



import unittest
import ircstates, irctokens
from ircstates.server import WHO_TYPE

class WHOTest(unittest.TestCase):
    def test_who(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]
        server.parse_tokens(irctokens.tokenise(
            "352 * #chan user host server nickname * :0 real"))

        self.assertEqual(user.username, "user")
        self.assertEqual(user.hostname, "host")
        self.assertEqual(user.realname, "real")
        self.assertEqual(user.server,   "server")
        self.assertIsNone(user.away)

        self.assertEqual(server.username, user.username)
        self.assertEqual(server.hostname, user.hostname)
        self.assertEqual(server.realname, user.realname)
        self.assertEqual(server.server,   user.server)
        self.assertIsNone(server.away)

        server.parse_tokens(irctokens.tokenise(
            "352 * #chan user host server nickname G* :0 real"))
        self.assertEqual(user.away,   "")
        self.assertEqual(server.away, "")

    def test_whox(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]
        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user 1.2.3.4 host server nickname * account :real"))

        self.assertEqual(user.username, "user")
        self.assertEqual(user.hostname, "host")
        self.assertEqual(user.realname, "real")
        self.assertEqual(user.account,  "account")
        self.assertEqual(user.server,   "server")
        self.assertIsNone(user.away)
        self.assertEqual(user.ip,       "1.2.3.4")

        self.assertEqual(server.username, user.username)
        self.assertEqual(server.hostname, user.hostname)
        self.assertEqual(server.realname, user.realname)
        self.assertEqual(server.account,  user.account)
        self.assertEqual(server.server,   user.server)
        self.assertIsNone(server.away)
        self.assertEqual(server.ip,       user.ip)

        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user realip host server nickname G account :real"))
        self.assertEqual(user.away,   "")
        self.assertEqual(server.away, "")

    def test_whox_no_account(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        user = server.users["nickname"]
        user.account   = "account"
        server.account = "account"

        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user realip host server nickname * 0 :real"))

        self.assertEqual(user.account,   None)
        self.assertEqual(server.account, user.account)

    def test_whox_ipv6(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        user = server.users["nickname"]

        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user 0::1 host server nickname * 0 :real"))
        self.assertEqual(user.ip, "::1")

        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user 00::2 host server nickname * 0 :real"))
        self.assertEqual(user.ip, "::2")

        server.parse_tokens(irctokens.tokenise(
            f"354 * {WHO_TYPE} user fd00:0:0:0::1 host server nickname * 0 :real"))
        self.assertEqual(user.ip, "fd00::1")



import unittest
import ircstates, irctokens

class SASLTestAccount(unittest.TestCase):
    def test_loggedin(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("900 * nick!user@host account *"))

        self.assertEqual(server.nickname, "nick")
        self.assertEqual(server.username, "user")
        self.assertEqual(server.hostname, "host")
        self.assertEqual(server.account,  "account")

    def test_loggedout(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("900 * nick!user@host account *"))
        server.parse_tokens(irctokens.tokenise("901 * nick1!user1@host1 *"))

        self.assertEqual(server.nickname, "nick1")
        self.assertEqual(server.username, "user1")
        self.assertEqual(server.hostname, "host1")
        self.assertEqual(server.account,  None)



from typing   import Dict, List, Optional, Set
from pendulum import DateTime

from .channel_user import ChannelUser
from .names        import Name

class Channel(object):
    def __init__(self, name: Name):
        self._name = name

        self.users: Dict[str, ChannelUser] = {}

        self.topic:        Optional[str]      = None
        self.topic_setter: Optional[str]      = None
        self.topic_time:   Optional[DateTime] = None

        self.created:      Optional[DateTime] = None

        self._list_modes_temp: Dict[str, List[str]]     = {}
        self.list_modes:       Dict[str, List[str]]     = {}
        self.modes:            Dict[str, Optional[str]] = {}

    def __repr__(self) -> str:
        return f"Channel(name={self.name!r})"

    def get_name(self) -> Name:
        return self._name
    @property
    def name(self) -> str:
        return self._name.normal
    @property
    def name_lower(self) -> str:
        return self._name.folded

    def add_mode(self,
            char: str,
            param: Optional[str],
            list_mode: bool):
        if list_mode:
            if param is not None:
                if not char in self.list_modes:
                    self.list_modes[char] = []
                if not param in self.list_modes[char]:
                    self.list_modes[char].append(param)
        else:
            self.modes[char] = param

    def remove_mode(self,
            char: str,
            param: Optional[str]):
        if char in self.list_modes:
            if (param is not None and
                    param in self.list_modes[char]):
                self.list_modes[char].remove(param)
        elif char in self.modes:
            del self.modes[char]



from typing import List
from .names import Name

class ChannelUser(object):
    def __init__(self,
            nickname: Name,
            channel_name: Name):
        self._nickname     = nickname
        self._channel_name = channel_name

        self.modes: List[str] = []

    def __repr__(self) -> str:
        outs: List[str] = [self.channel, self.nickname]
        if self.modes:
            outs.append(f"+{''.join(self.modes)}")
        return f"ChannelUser({' '.join(outs)})"

    @property
    def nickname(self) -> str:
        return self._nickname.normal
    @property
    def nickname_lower(self) -> str:
        return self._nickname.folded

    @property
    def channel(self) -> str:
        return self._channel_name.normal



import unittest
import ircstates, irctokens

class EmitTest(unittest.TestCase):
    def test_join(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        emit = server.parse_tokens(
            irctokens.tokenise(":nickname JOIN #chan"))

        self.assertEqual(emit.command, "JOIN")
        self.assertEqual(emit.self,    True)
        self.assertEqual(emit.user,    server.users["nickname"])
        self.assertEqual(emit.channel, server.channels["#chan"])

        emit = server.parse_tokens(
            irctokens.tokenise(":other JOIN #chan"))
        self.assertIsNotNone(emit)
        self.assertEqual(emit.command, "JOIN")
        self.assertEqual(emit.self,    None)
        self.assertEqual(emit.user,    server.users["other"])
        self.assertEqual(emit.channel, server.channels["#chan"])

    def test_privmsg(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        emit = server.parse_tokens(
            irctokens.tokenise(":nickname PRIVMSG #chan :hello"))
        self.assertIsNotNone(emit)
        self.assertEqual(emit.command,     "PRIVMSG")
        self.assertEqual(emit.text,        "hello")
        self.assertEqual(emit.self_source, True)
        self.assertEqual(emit.user,        server.users["nickname"])
        self.assertEqual(emit.channel,     server.channels["#chan"])

        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        emit = server.parse_tokens(
            irctokens.tokenise(":other PRIVMSG #chan :hello2"))
        self.assertIsNotNone(emit)
        self.assertEqual(emit.command,     "PRIVMSG")
        self.assertEqual(emit.text,        "hello2")
        self.assertEqual(emit.self_source, None)
        self.assertEqual(emit.user,        server.users["other"])
        self.assertEqual(emit.channel,     server.channels["#chan"])

    def test_privmsg_nojoin(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))

        emit = server.parse_tokens(
            irctokens.tokenise(":other PRIVMSG #chan :hello"))

        self.assertIsNotNone(emit)
        self.assertEqual(emit.command,     "PRIVMSG")
        self.assertEqual(emit.text,        "hello")
        self.assertEqual(emit.self_source, None)
        self.assertIsNotNone(emit.user)
        channel = server.channels["#chan"]
        self.assertEqual(emit.channel,     channel)

    def test_kick(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]
        channel = server.channels["#chan"]
        server.parse_tokens(irctokens.tokenise(":other JOIN #chan"))
        user_other = server.users["other"]
        emit = server.parse_tokens(
            irctokens.tokenise(":nickname KICK #chan other :reason"))

        self.assertIsNotNone(emit)
        self.assertEqual(emit.command,     "KICK")
        self.assertEqual(emit.text,        "reason")
        self.assertEqual(emit.self_source, True)
        self.assertEqual(emit.user_source, user)
        self.assertEqual(emit.user_target, user_other)
        self.assertEqual(emit.channel,     channel)

    def test_mode_self(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        emit = server.parse_tokens(
            irctokens.tokenise("MODE nickname x+i-i+wi-wi"))

        self.assertIsNotNone(emit)
        self.assertEqual(emit.command, "MODE")
        self.assertTrue(emit.self_target)
        self.assertEqual(emit.tokens,
            ["+x", "+i", "-i", "+w", "+i", "-w", "-i"])

    def test_mode_channel(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        channel = server.channels["#chan"]
        emit = server.parse_tokens(
            irctokens.tokenise(":server MODE #chan +im-m+b-k asd!*@* key"))

        self.assertIsNotNone(emit)
        self.assertEqual(emit.command, "MODE")
        self.assertEqual(emit.channel, channel)
        self.assertEqual(emit.tokens,
            ["+i", "+m", "-m", "+b asd!*@*", "-k key"])



import unittest
import ircstates, irctokens

class CapTestLS(unittest.TestCase):
    def test_one_line(self):
        server = ircstates.Server("test")
        self.assertFalse(server.has_cap)
        self.assertEqual(server.available_caps, {})
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b"))
        self.assertEqual(server.available_caps, {"a": "", "b": ""})

    def test_two_lines(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS * :a b"))
        self.assertEqual(server.available_caps, {})
        server.parse_tokens(irctokens.tokenise("CAP * LS :c"))
        self.assertEqual(server.available_caps, {"a": "", "b": "", "c": ""})

    def test_values(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b= c=1"))
        self.assertEqual(server.available_caps, {"a": "", "b": "", "c": "1"})

class CapTestACK(unittest.TestCase):
    def test_one_line(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK :a b"))
        self.assertEqual(server.agreed_caps, ["a", "b"])

    def test_two_lines(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b c"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK * :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK :c"))
        self.assertEqual(server.agreed_caps, ["a", "b", "c"])

    def test_not_ls(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS a"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK b"))
        self.assertEqual(server.agreed_caps, [])

class CapTestNEW(unittest.TestCase):
    def test_no_ls(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * NEW :a"))
        self.assertEqual(server.available_caps, {"a": ""})

    def test_one(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a"))
        server.parse_tokens(irctokens.tokenise("CAP * NEW :b"))
        self.assertEqual(server.available_caps, {"a": "", "b": ""})

    def test_two(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a"))
        server.parse_tokens(irctokens.tokenise("CAP * NEW :b c"))
        self.assertEqual(server.available_caps, {"a": "", "b": "", "c": ""})

class CapTestDEL(unittest.TestCase):
    def test_not_acked(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * DEL a"))

    def test_one_ls(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK :a"))
        server.parse_tokens(irctokens.tokenise("CAP * DEL :a"))
        self.assertEqual(server.available_caps, {})
        self.assertEqual(server.agreed_caps, [])

    def test_two_ls(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * DEL :a"))
        self.assertEqual(server.available_caps, {"b": ""})
        self.assertEqual(server.agreed_caps, ["b"])

    def test_two_del(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("CAP * LS :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * ACK :a b"))
        server.parse_tokens(irctokens.tokenise("CAP * DEL :a b"))
        self.assertEqual(server.available_caps, {})
        self.assertEqual(server.agreed_caps, [])





RPL_WELCOME     = "001"
RPL_ISUPPORT    = "005"
RPL_MOTD        = "372"
RPL_MOTDSTART   = "375"
RPL_UMODEIS     = "221"
RPL_VISIBLEHOST = "396"
RPL_TRYAGAIN    = "263"

ERR_NOSUCHNICK   = "401"
ERR_NOSUCHSERVER = "402"

RPL_CHANNELMODEIS = "324"
RPL_CREATIONTIME  = "329"
RPL_TOPIC         = "332"
RPL_TOPICWHOTIME  = "333"

RPL_WHOREPLY   = "352"
RPL_WHOSPCRPL  = "354"
RPL_ENDOFWHO   = "315"
RPL_NAMREPLY   = "353"
RPL_ENDOFNAMES = "366"

RPL_WHOWASUSER  = "314"
RPL_ENDOFWHOWAS = "369"

RPL_BANLIST        = "367"
RPL_ENDOFBANLIST   = "368"
RPL_QUIETLIST      = "728"
RPL_ENDOFQUIETLIST = "729"

RPL_LOGGEDIN    = "900"
RPL_LOGGEDOUT   = "901"
RPL_SASLSUCCESS = "903"
ERR_SASLFAIL    = "904"
ERR_SASLTOOLONG = "905"
ERR_SASLABORTED = "906"
ERR_SASLALREADY = "907"
RPL_SASLMECHS   = "908"

RPL_WHOISUSER     = "311"
RPL_WHOISSERVER   = "312"
RPL_WHOISOPERATOR = "313"
RPL_WHOISIDLE     = "317"
RPL_WHOISCHANNELS = "319"
RPL_WHOISACCOUNT  = "330"
RPL_WHOISHOST     = "378"
RPL_WHOISMODES    = "379"
RPL_WHOISSECURE   = "671"
RPL_AWAY          = "301"
RPL_ENDOFWHOIS    = "318"

ERR_ERRONEUSNICKNAME = "432"
ERR_NICKNAMEINUSE    = "433"
ERR_BANNICKCHANGE    = "435"
ERR_UNAVAILRESOURCE  = "437"
ERR_NICKTOOFAST      = "438"
ERR_CANTCHANGENICK   = "447"

ERR_NOSUCHCHANNEL   = "403"
ERR_TOOMANYCHANNELS = "405"
ERR_USERONCHANNEL   = "443"
ERR_LINKCHANNEL     = "470"
ERR_BADCHANNAME     = "479"
ERR_BADCHANNEL      = "926"


ERR_BANNEDFROMCHAN = "474"
ERR_INVITEONLYCHAN = "473"
ERR_BADCHANNELKEY  = "475"
ERR_CHANNELISFULL  = "471"
ERR_NEEDREGGEDNICK = "477"
ERR_THROTTLE       = "480"



import unittest
import ircstates, irctokens

class MOTDTest(unittest.TestCase):
    def test(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise("375 * :start of motd"))
        server.parse_tokens(irctokens.tokenise("372 * :first line of motd"))
        server.parse_tokens(irctokens.tokenise("372 * :second line of motd"))
        self.assertEqual(server.motd, [
            "start of motd",
            "first line of motd",
            "second line of motd"])



from .server       import Server, ServerDisconnectedException
from .user         import User
from .channel      import Channel
from .channel_user import ChannelUser
from .casemap      import casefold
from .emit         import *



from typing import Optional, Set
from .names import Name

class User(object):
    def __init__(self, nickname: Name):
        self._nickname = nickname

        self.username: Optional[str] = None
        self.hostname: Optional[str] = None
        self.realname: Optional[str] = None
        self.account:  Optional[str] = None
        self.server:   Optional[str] = None
        self.away:     Optional[str] = None
        self.ip:       Optional[str] = None
        self.channels: Set[str]      = set([])

    def __repr__(self) -> str:
        return f"User(nickname={self.nickname!r})"

    def get_name(self) -> Name:
        return self._nickname
    @property
    def nickname(self) -> str:
        return self._nickname.normal
    @property
    def nickname_lower(self) -> str:
        return self._nickname.folded

    def change_nickname(self,
            normal: str,
            folded: str):
        self._nickname.normal = normal
        self._nickname.folded = folded

    def hostmask(self) -> str:
        hostmask = self.nickname
        if self.username is not None:
            hostmask += f"!{self.username}"
        if self.hostname is not None:
            hostmask += f"@{self.hostname}"
        return hostmask

    def userhost(self) -> Optional[str]:
        if (self.username is not None and
                self.hostname is not None):
            return f"{self.username}@{self.hostname}"
        else:
            return None



import unittest
import ircstates, irctokens

class CaseMapTestMethod(unittest.TestCase):
    def test_rfc1459(self):
        lower = ircstates.casefold("rfc1459", "ÀTEST[]^\\")
        self.assertEqual(lower, "Àtest{}~|")

    def test_ascii(self):
        lower = ircstates.casefold("ascii", "ÀTEST[]~\\")
        self.assertEqual(lower, "Àtest[]~\\")

class CaseMapTestCommands(unittest.TestCase):
    def test_join(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":Nickname JOIN #Chan"))
        server.parse_tokens(irctokens.tokenise(":Other JOIN #Chan"))
        self.assertIn("nickname", server.users)
        self.assertNotIn("Nickname", server.users)
        self.assertIn("other", server.users)
        self.assertNotIn("Other", server.users)
        self.assertIn("#chan", server.channels)
        self.assertNotIn("#Chan", server.channels)

        channel = server.channels["#chan"]
        self.assertEqual(channel.name, "#Chan")

    def test_nick(self):
        server = ircstates.Server("test")
        server.parse_tokens(irctokens.tokenise("001 nickname *"))
        server.parse_tokens(irctokens.tokenise(":nickname JOIN #chan"))
        user = server.users["nickname"]
        server.parse_tokens(irctokens.tokenise(":nickname NICK NewNickname"))
        self.assertEqual(len(server.users), 1)
        self.assertIn("newnickname", server.users)
        self.assertEqual(user.nickname, "NewNickname")
        self.assertEqual(user.nickname_lower, "newnickname")
        self.assertEqual(server.nickname, "NewNickname")
        self.assertEqual(server.nickname_lower, "newnickname")



from typing import List, Optional

class ChanModes(object):
    def __init__(self,
            a_modes: List[str],
            b_modes: List[str],
            c_modes: List[str],
            d_modes: List[str]):
        self.a_modes = a_modes
        self.b_modes = b_modes
        self.c_modes = c_modes
        self.d_modes = d_modes

class Prefix(object):
    def __init__(self,
            modes:    List[str],
            prefixes: List[str]):
        self.modes    = modes
        self.prefixes = prefixes

    def from_mode(self, mode: str) -> Optional[str]:
        if mode in self.modes:
            return self.prefixes[self.modes.index(mode)]
        return None
    def from_prefix(self, prefix: str) -> Optional[str]:
        if prefix in self.prefixes:
            return self.modes[self.prefixes.index(prefix)]
        return None
